`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: PierceATronics
// Engineer: David Pierce Walker-Howell
// 
// Create Date: 02/29/2020 02:24:08 PM
// Design Name: 
// Module Name: IIR_filter_biquad_df1
// Project Name: 
// Target Devices: 
// Tool Versions: 
// Description: 
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////

/*
     This module is computes the IIR response to an input signal. The form of the biquad
     filter is given by.
                b_0 + b_1*z^(-1) + b_2*z^(-2)
     H(z) = --------------------------------------
                 1 + a_1*z^(-1) + a_2*z^(-2)
*/
     
module IIR_filter_biquad_df1 #(parameter BIT_WIDTH=16, F=7) 
    (input SYS_CLK, SAMPLE_CLK, RST, LOAD,                         //System Clock, Sample Clock, reset, and load signal
     input [2:0] coeff_addr,                //Address of coefficient to load 3'b000 => a_1, 3'b001 => a_2, 3'b010 => b_0, etc.
     input signed [BIT_WIDTH-1:0] coeff,    //The coefficient to load
     input signed [BIT_WIDTH-1:0] x_in,     //Signed Input data
     output signed [BIT_WIDTH-1:0] y_out,
     output LOAD_OKAY); 
     
    //a[0] -> a_1, a[1] -> a_2
    //b[0] -> b_0, b[1] -> b_1, b[2] -> b_2 
    reg signed [BIT_WIDTH-1:0] b[2:0], a[1:0]; //coefficients for the IIR filter  
    reg signed [BIT_WIDTH-1:0] x [1:0]; //Stores the input signal i.e x[n], x[n-1], x[n-2]
    reg signed [BIT_WIDTH-1:0] y [1:0]; //Stores the output signal i.e y[n], y[n-1], y[n-2]
    //reg signed [2*BIT_WIDTH-1:0] temp [4:0]; //Temporarily stores each multiplication
	 reg signed [2*BIT_WIDTH-1:0] temp_0, temp_1, temp_2, temp_3, temp_4; //Temporarily stores each multiplication
    reg signed [2*BIT_WIDTH-1:0] accumulator;
    reg prev_sample_clk, start_filter_comp, accumulation_state; //Start filter computation flag
    reg [3:0] accumulation_counter;
    
	 
    //Initialize the coefficients
    initial begin
        a[0] = 0; 
        a[1] = 0; 
        b[0] = (1 << 7);
        b[1] = 0; 
        b[2] = 0; 
        x[0] = 0; x[1] = 0;
        y[0] = 0; y[1] = 0;
        accumulation_counter = 0;
    end
    
    //Loading new coefficients is only okay when the filter is not performing a computation.
    assign LOAD_OKAY = (start_filter_comp | accumulation_state) ? (1'b0) : (1'b1);
    
    //LOADING STATE
    always @(posedge SYS_CLK) begin
        if(RST) begin
            //Set the coefficients so the filter just passes the signal
            a[0] <= 0; 
            a[1] <= 0; 
            b[0] <= (1 << 7);
            b[1] <= 0; 
            b[2] <= 0;
        end
        else if(LOAD_OKAY & LOAD) begin
            case(coeff_addr)
                3'b000:
                    a[0] <= coeff;
                3'b001:
                    a[1] <= coeff;
                3'b010:
                    b[0] <= coeff;
                3'b011:
                    b[1] <= coeff;
                3'b100:
                    b[2] <= coeff;
            endcase
        end
    end
     
    //Detect a positive edge of the sampling clock. This sets a flag to start a single filter computation.
    always @(posedge SYS_CLK) begin
        
        if(RST) begin
            start_filter_comp <= 0;
        end
        else if( (prev_sample_clk == 1'b0) & (SAMPLE_CLK == 1'b1)) 
            start_filter_comp <= 1; 
        else
            start_filter_comp <= 0;
        
        prev_sample_clk <= SAMPLE_CLK;
    end
    
    //Perform filter computation
    always @(posedge SYS_CLK) begin
        
        if(RST) begin  //On reset
             x[0] <= 0; x[1] <= 0;
             y[0] <= 0; y[1] <= 0;
				 temp_0 <= 0; temp_1 <= 0; temp_2 <= 0; temp_3 <= 0; temp_4 <= 0; 
             accumulation_state <= 0;
             accumulation_counter <= 0;
             accumulator <= 0;
        end
        
        //Start the filter computation by computing all the multiplications of coefficients
        else if(start_filter_comp & ~accumulation_state) begin
				/*
            temp[0] <= a[0] * y[0];
            temp[1] <= a[1] * y[1];
            temp[2] <= b[0] * x_in;
            temp[3] <= b[1] * x[0];
            temp[4] <= b[2] * x[1];
            */
				temp_0 <= a[0] * y[0];
            temp_1 <= a[1] * y[1];
            temp_2 <= b[0] * x_in;
            temp_3 <= b[1] * x[0];
            temp_4 <= b[2] * x[1];
            //Shift over signal data
            y[1] <= y[0];
            x[0] <= x_in;
            x[1] <= x[0];
            
            //Make next state the accumulation state
            accumulation_state <= 1;
            accumulation_counter <= 0;
            accumulator <= 0;
        end
        
        //accumulation state. There are 5 additions to be computed
        else if(accumulation_state) begin
            /*
            if(accumulation_counter > 4) begin
                accumulation_counter <= 0;
                accumulator <= 0;
                accumulation_state <= 0;
                //Since the accumulation is 2*BIT_WIDTH long with decimal at 2*F.
                //The result needs to be taken at BIT_WIDTH long with decimal at F.
                y[0] <= accumulator[BIT_WIDTH + F - 1:F];
            end
            else begin
                accumulation_counter <= accumulation_counter + 1;
					 
                if(accumulation_counter < 2)
                    accumulator <= accumulator - temp[accumulation_counter];
                else
                    accumulator <= accumulator + temp[accumulation_counter];
                accumulation_state <= 1;
            end
				*/
				temp_0 <= temp_0; temp_1 <= temp_1; temp_2 <= temp_2; temp_3 <= temp_3; temp_4 <= temp_4;
				case(accumulation_counter)
					0: begin
						accumulator <= accumulator - temp_0;
						accumulation_counter <= 1;
						accumulation_state <= 1;
					end
					1: begin
						accumulator <= accumulator - temp_1;
						accumulation_counter <= 2;
						accumulation_state <= 1;
					end
					2: begin
						accumulator <= accumulator + temp_2;
						accumulation_counter <= 3;
						accumulation_state <= 1;
					end
					3: begin
						accumulator <= accumulator + temp_3;
						accumulation_counter <= 4;
						accumulation_state <= 1;
					end
					4: begin
						accumulator <= accumulator + temp_4;
						accumulation_counter <= 5;
						accumulation_state <= 1;
					end
					
					5: begin
						accumulation_counter <= 0;
						accumulator <= accumulator;
						accumulation_state <= 0;
						//Since the accumulation is 2*BIT_WIDTH long with decimal at 2*F.
						//The result needs to be taken at BIT_WIDTH long with decimal at F.
						y[0] <= accumulator[BIT_WIDTH + F - 1:F];
					end
					
					default: begin
						accumulator <= accumulator;
						accumulation_counter <= 0;
						accumulation_state <= 0;
					end
				
				endcase
        end
        
		  else begin
				 x[0] <= x[0]; x[1] <= x[1];
             y[0] <= y[0]; y[1] <= y[1];
				 temp_0 <= 0; temp_1 <= 0; temp_2 <= 0; temp_3 <= 0; temp_4 <= 0; 
             accumulation_state <= accumulation_state;
             accumulation_counter <= accumulation_counter;
             accumulator <= accumulator;
		  end
        
		  
    end
    //Assign the output
    assign y_out = y[0];
    
    
endmodule
