`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: PierceATronics
// Engineer: David Pierce Walker-Howell
// 
// Create Date: 03/25/2020 02:46:51 PM
// Design Name: 
// Module Name: coefficient_controller
// Project Name: Parametric_EQ_Sim
// Target Devices: 
// Tool Versions: 
// Description: 
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////

/*
This modules function is to control the storing and loading of IIR biquad filter
coefficients. Once the coefficients are received via UART, the coefficient controller
will store them into a memory block, and then once the filters are in a stable state 
for changing coefficients, the coefficient controller will load the coefficients into the
respective filter.
*/
module coefficient_controller (
    input SYS_CLK, RST, LOAD_EN,
    //input [1:0] LOAD_OKAY_N //FOR MULTIPLE FILTERS USE: The signal for each filter on when is good to load
    input LOAD_OKAY,
    input WRITE_EN,
    input [2:0] write_addr,
    input signed [15:0] coeff_in,
    output reg LOAD, //Load flag to tell filters that coefficients are going to be loaded.
    output signed [15:0] coeff_out, //The coefficient to load to the filter
    output reg [2:0] coeff_addr
    );
    
    reg prev_load_sig; //Holds the previous LOAD_OKAY signal sampled
    reg start_coeff_load; //Flag to start loading coefficients
    reg state; //Either IDLE state or LOAD_1 for loading filter 1.
    wire signed [15:0] coeff_out_temp;
    reg [2:0] coeff_addr_temp; //Used so it can be registered to line up with coeff_out value.
    
    //Instantiate the memory block were coefficients are stored.
    simple_dual_port_mem #(.ADDR_WIDTH(3), .DATA_SIZE(16)) coeff_mem (.CLK(SYS_CLK), .RST(RST), .WRITE_EN(WRITE_EN),
                    .data_in(coeff_in), .write_addr(write_addr), .read_addr(coeff_addr_temp), .data_out(coeff_out_temp));
    
    //Coefficient to load to filter when LOAD flag is high
    assign coeff_out = (LOAD) ? (coeff_out_temp) : 0;
    
    //Detect when the load okay has a rising edge. This starts the coefficient loading process.
    always @(posedge SYS_CLK) begin
        prev_load_sig <= LOAD_OKAY;
        if(RST)
            start_coeff_load <= 1'b0;
        else if( (prev_load_sig == 1'b0) & (LOAD_OKAY == 1'b1))
            start_coeff_load <= 1'b1;
        else
            start_coeff_load <= 1'b0;
    end
    
    //Control of loading filter coefficients into the filter.
    always @(posedge SYS_CLK) begin
        if(RST | ~LOAD_EN) begin
            state <= 0; //IDLE state
            coeff_addr_temp <= 0;
            coeff_addr <= 0;
            LOAD <= 0;
        end
        else begin
            case(state)
         
                1'b0: begin //IDLE state
                    coeff_addr_temp <= 0;
                    if(start_coeff_load) begin //Go to load coefficients into the filter
                        state <= 1'b1;
                        LOAD <= 1'b1; 
                    end
                    else begin
                        state <= 1'b0;  //Stay in IDLE state
                        LOAD <= 1'b0;
                    end
                end
                
                1'b1: begin
                    if(coeff_addr_temp < 5) begin
                        LOAD <= 1'b1;
                        state <= 1'b1; //Stay in loading state
                        coeff_addr_temp <= (coeff_addr_temp == 4) ? 0 : coeff_addr_temp + 1;  //coeff_addr is used to access coefficients from the memory
                    end
                    else begin
                        LOAD <= 1'b0;
                        state <= 1'b0; //Go back to idle state
                        coeff_addr_temp <= 0;
                    end
                end
                
            endcase
 
            //Register the coefficient address so coeff_addr lines up with corresponding coeff_out
            coeff_addr <= coeff_addr_temp;
        
        end
        
    end
    
    /*FOR MULTIPLE FILTERS CASE
    wire LOAD_OKAY; //Load okay signal for current current filter being loaded
    assign LOAD_OKAY = LOAD_OKAY_N[filt_num];
    */
    /* FOR FUTURE USE WITH MULTIPLE FILTERS
    parameter IDLE = 0;
    parameter NUM_FILTS = 2;
    reg [1:0] state;
    reg [1:0]prev_filt; // The previous filter that was loaded with coefficients
    wire filt_num;
    
    //State machine that loads coefficients into given filter.
    always @(posedge SYS_CLK) begin
        
        case(state)
            IDLE: begin//when the start coeff
                prev_filt <= prev_filt;
                if(start_coeff_load) begin
                    if(prev_filt < NUM_FILTS)
                        state <= prev_filt + 1;
                    else
                        state <= 1;
                end
                else
                    state <= IDLE;
            end
            2'b01: begin //loading for filter 1
                prev_filt <= 1'b1;
                //Perform loading
                
                
            end 
            2'b10: begin
                prev_filt <= 2'b10;
            
            end
            
            default: begin
                prev_filt <= 2'b10;
                state <= IDLE;
            end
        endcase
    end
    */
endmodule
