`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: PierceATronics
// Engineer: David Pierce Walker-Howell
// 
// Create Date: 11/03/2019 11:58:10 AM
// Design Name: 
// Module Name: digilent_PMOD_DA2
// Project Name: COMPE-470L Final Project: FIR_Filters
// Target Devices: Minized
// Tool Versions: Vivado 2018.3
// Description: This module provides the communication control to the digilent PMOD DA2
//              2-channel, 12-bit Digital-to-Analog converter. It uses an SPI like interface.
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////


module digilent_PMOD_DA2(input SYS_CLK, RST, WE, 
                         input [11:0] DATA_1_IN, DATA_2_IN,
                         input [1:0] MODE,
                         output reg WC, 
                         output reg CS, SCLK, DATA_1_OUT, DATA_2_OUT);
/*
    Input Parameters:
        SYS_CLK: The system clock
        RST: The reset, synchronous active high.
        WE: Write enable, active high. If low, then device is kept in ideal state. 
            This means CS, SCLK, DATA_1_OUT, and DATA_2_OUT will be low.
        DATA_1_IN: 12 bit data to send on channel 1.
        DATA_2_IN: 12 bit data to send on channel 2.
        MODE: The output mode operation of device.
                00: Normal Operation
                01: 1k Ohm to GND
                10: 100K Ohm to GND
                11: HIGH IMPEDANCE
    Output Parameters:
        WC: Write complete. Driven high once a write is complete, goes low
            when a new sending sequence starts.
        CS: Chip Select. When driven from high to low, that lets the PMOD device know 
                         to user is going to send a series of 16 clock pulses on the SCLK
                         line. When in ideal state, this should be held low to make sure there
                         is low power consumption.
        SCLK: Serial Clock Line. At falling edges the device will read the data. 16 pulses per 
              sequence.
        DATA_1_OUT: The serial data sent to channel 1, this will be held low during ideal state.
        DATA_2_OUT: The serial data sent to channel 2, this will be held low during ideal state.              
*/
    localparam IDLE_STATE = 0, INITIALIZE_SEQUENCE = 1, SEND_SEQUENCE = 2;
    localparam SCLK_COUNTS = 1;
    reg [1:0] state;
    reg [8:0] shift_counter; //Keeps count of what serial data is being sent in the sequency
    reg [15:0] DATA_1_REG, DATA_2_REG;
    reg [11:0] sclk_clock_divider_counter;
    
    always @(posedge SYS_CLK or posedge RST) begin
        if(RST) begin //Active high reset
            state <= IDLE_STATE; //Go into idle state;
            WC <= 0; //Write complete initially false
            CS <= 0; //Ideal the chip select low to save power. Same for SCLK and output data
            SCLK <= 0; 
            DATA_1_OUT <= 0; DATA_2_OUT <= 0; 
            DATA_1_REG <= 0; DATA_2_REG <= 0;
            shift_counter <= 0; sclk_clock_divider_counter <= 0;
        end
        
        else begin
            //Finite State Machine to control the device
            case(state)
            
                IDLE_STATE: begin
                    shift_counter <= 0; sclk_clock_divider_counter <= 0;
                    if(WE) begin //If write enable is high, start the writing sequency
                        state <= INITIALIZE_SEQUENCE;
                        WC <= 0;
                        CS <= 1; //Pull the chip select high before writing a sequence.
                        SCLK <= 1;
                        //SHIFT IN THE DATA
                        {DATA_1_REG[13:12], DATA_1_REG[11:0]}  <= {MODE, DATA_1_IN};
                        {DATA_2_REG[13:12], DATA_2_REG[11:0]}  <= {MODE, DATA_2_IN};
                        DATA_1_OUT <= DATA_1_OUT; DATA_1_OUT <= DATA_1_OUT;
                    end
                    else begin
                        state <= IDLE_STATE; //Go into idle state;
                        WC <= 0; //Write complete initially false
                        CS <= 0; //Ideal the chip select low to save power. Same for SCLK and output data
                        SCLK <= 0; 
                        DATA_1_OUT <= 0; DATA_2_OUT <= 0; 
                        DATA_1_REG <= 0; DATA_2_REG <= 0;
                    end
                 end
                
                INITIALIZE_SEQUENCE: begin
                    shift_counter <= 0; sclk_clock_divider_counter <= 0;
                    state <= SEND_SEQUENCE;
                    CS <= 0; //CS will now go from high to low.
                    DATA_1_REG <= DATA_1_REG; DATA_2_REG <= DATA_2_REG;
                    DATA_1_OUT <= 0;
                    DATA_2_OUT <= 0;
                    SCLK <= 1;
                    WC <= 0;
                end
                
                SEND_SEQUENCE: begin
                    
                    if(shift_counter <= 15) begin
                    
                        CS <= 0; //Hold the CS line low while sending
                        DATA_1_REG <= DATA_1_REG; DATA_2_REG <= DATA_2_REG; // Hold the registered data   
                        //When the sclk goes high, shift next data bit to send
                        DATA_1_OUT <= (SCLK) ? (DATA_1_REG[15 - shift_counter]) : (DATA_1_OUT);
                        DATA_2_OUT <= (SCLK) ? (DATA_2_REG[15 - shift_counter]) : (DATA_2_OUT);
                        
                        //SCLK CLOCK COUNTER
                        //--------------------------------------------------------------------------
                        if(sclk_clock_divider_counter < SCLK_COUNTS) begin
                            sclk_clock_divider_counter <= sclk_clock_divider_counter + 1;
                            shift_counter <= shift_counter;
                        end
                        else begin
                            shift_counter <= (SCLK) ? (shift_counter + 1) : shift_counter;
                            SCLK <= ~SCLK;
                            sclk_clock_divider_counter <= 0;
                        end 
                        //----------------------------------------------------------------------------
                    end
                    
                    //SENDING COMPLETE
                    else begin
                        shift_counter <= 0; sclk_clock_divider_counter <= 0;
                        if(WE) begin //If write enable is high, start the writing sequency
                            state <= INITIALIZE_SEQUENCE;
                            WC <= 1;
                            CS <= 1; //Pull the chip select high before writing a sequence.
                            SCLK <= 1;
                            //SHIFT IN THE DATA
                            {DATA_1_REG[13:12], DATA_1_REG[11:0]}  <= {MODE, DATA_1_IN};
                            {DATA_2_REG[13:12], DATA_2_REG[11:0]}  <= {MODE, DATA_2_IN};
                            DATA_1_OUT <= DATA_1_OUT; DATA_1_OUT <= DATA_1_OUT;
                        end
                        else begin
                            state <= IDLE_STATE; //Go into idle state;
                            WC <= 1; //Write complete initially false
                            CS <= 0; //Ideal the chip select low to save power. Same for SCLK and output data
                            SCLK <= 0; 
                            DATA_1_OUT <= 0; DATA_2_OUT <= 0; 
                            DATA_1_REG <= 0; DATA_2_REG <= 0;
                        end
                    end
                end
                
            
            endcase
        end
    end
endmodule
