`timescale 1ns / 1ps

module receive(input CLK, RX_IN, input RXE,
                output reg RXC, output reg [7:0] UDRRX);
     /*
    *This module is the receive (RX) state machine for UART
    *It receives an 8-bit data frame with an additional start and stop bit.
    * RX=> Receive Enable
    *CLK=> A clock with frequency 16 times faster than the BAUD rate.
    *RX_IN=> The input pin where RX data is received from.
    *RXE=> Receive Enable 
    *RXC=> An output flag to signify a receive  complete.
    *UDRRX => The data register holding the received data frame.
    */
    localparam DISABLED = 0, IDLE = 1, GLITCH_CHECK = 2, RECEIVING = 3;
    reg [1:0] state;
    reg [8:0] glitch_counter, shift_counter, read_bit_counter;
    reg [7:0] UDRRX_buffer;
    
    initial begin
        state <= DISABLED;
        glitch_counter <= 0;
        shift_counter <= 0;
        read_bit_counter <= 0; 
        RXC <= 0;
        UDRRX_buffer <= 0;
    end
    
    //NEXT STATE LOGIC
    always @(posedge CLK) begin
        
        //RECEIVE NOT ENABLED
        if(!RXE) begin
            state <= DISABLED;
            glitch_counter <= 0;
            shift_counter <= 0;
            read_bit_counter <= 0; 
            RXC <= 0;
            UDRRX_buffer <= 0;
        end
        
        else begin
            case(state)
                //
                DISABLED: begin
                    glitch_counter <= 0;
                    shift_counter <= 0;
                    read_bit_counter <= 0;
                    RXC <= 0;
                    UDRRX_buffer <= 0;
                    //IF RX Enabled and RX_IN is 1 (idling) then go into idle state
                    if(RXE && RX_IN) state <= IDLE;
                    else state <= DISABLED;
                end
               
                //IDLING STATE
                IDLE: begin
                    shift_counter <= 0;
                    read_bit_counter <= 0;
                    RXC <= RXC;
                    UDRRX_buffer <= 0;
                    //If data goes from one to zero
                    if(!RX_IN) begin
                        state <= GLITCH_CHECK;
                        glitch_counter <= 1;
                    end
                    else begin
                        state <= IDLE;
                        glitch_counter <= 0;
                    end
                end
                
                //GLITCH_CHECK 
                //In the glitch check state, the receiver is making sure the start bit isn't mistaken for a 
                //zero glitch.
                GLITCH_CHECK: begin
                    shift_counter <= 0;
                    read_bit_counter <= 0;
                    RXC <= 0;
                    UDRRX <= UDRRX;
                    UDRRX_buffer <= UDRRX_buffer;
                    
                    if(glitch_counter < 8) begin
                        glitch_counter <= glitch_counter + 1;
                        state <= GLITCH_CHECK;
                    end
                    else begin
                        glitch_counter <= 0;
                        if(!RX_IN) state <= RECEIVING;
                        else state <= DISABLED;
                    end
                end
                
                //RECEIVING 8-bits of data and checking for valid stop bit
                RECEIVING: begin
                    glitch_counter <= 0;
                    //Reading in the 8-bit data from the RX in pin.
                    if(shift_counter < 8) begin
                        RXC <= 0; // receive not complete
                        UDRRX <= UDRRX;
                        state <= RECEIVING;
                        if(read_bit_counter < 15) begin
                            read_bit_counter <= read_bit_counter + 1;
                            shift_counter <= shift_counter;
                            UDRRX_buffer <= UDRRX_buffer;
                        end
                        else begin
                            read_bit_counter <= 0;
                            shift_counter <= shift_counter + 1;
                            UDRRX_buffer[shift_counter] <= RX_IN;
                        end
                    end
                    
                    //After the 8 data bits are read, check for the stop bit
                    else begin
                        shift_counter <= shift_counter;
                        UDRRX_buffer <= UDRRX_buffer;
                        
                        //Wait to read until in the middle of potential stop bit
                        if(read_bit_counter < 15) begin
                            read_bit_counter <= read_bit_counter + 1;
                            RXC <= 0;
                            state <= RECEIVING;
                        end
                        
                        //Once in the middle of the potential stop bit, check to see if
                        //it is a valid start bit (meaning LOGIC 1)
                        else begin
                            read_bit_counter <= 0;
                            //IF a valid stop bit, move the buffered data into the UDRRX register 
                            //and set the receive complete flag to true
                            if(RX_IN) begin
                                RXC <= 1;
                                UDRRX <= UDRRX_buffer;
                                state <= IDLE;
                            end
                            
                            //IF invalid stop bit, go into the disabled state and wait for the next
                            //logic 1 to appear to be able to go into the IDLE state.
                            else begin
                                RXC <= 0;
                                state <= DISABLED;
                                UDRRX <= UDRRX;
                            end
                        end
                    end
                end
            endcase
        end
    end
endmodule
