`timescale 1ns / 1ps
//////////////////////////////////////////////////////////////////////////////////
// Company: PierceATronics
// Engineer: David Pierce Walker-Howell
// 
// Create Date: 03/31/2020 02:44:17 PM
// Design Name: 
// Module Name: unpack_bytes
// Project Name: 
// Target Devices: 
// Tool Versions: 
// Description: 
// 
// Dependencies: 
// 
// Revision:
// Revision 0.01 - File Created
// Additional Comments:
// 
//////////////////////////////////////////////////////////////////////////////////


module unpack_bytes #(parameter NUM_FILTERS=1)
    (
    input SYS_CLK, RST,
    input [7:0] UDRRX,
    input RXC,
    output reg [7:0] filter_num,
    output reg [2:0] coeff_addr,
    output reg [15:0] coeff_out, 
    output reg WRITE_EN
    );
    
    localparam START_BYTE_1 = 0, START_BYTE_2 = 1, FILTER_NUM = 2, UNPACK_BYTES = 3, STOP_BYTE = 4, LOAD_STATE = 5;
    
    reg [2:0] state, prev_state;
    reg [3:0] unpack_bytes_counter; //There are 10 bytes to unpack
    reg [7:0] data; //Data read from the UART receive module. 
    reg at_msb;
    reg START_LOADING, LOADING_FINISHED; //The start of loading coefficients and the loading is finished.
    
    //Instantiate the memory to store the complete receive filter coefficients
    reg [15:0] coeff_mem [4:0]; //Holds the 5 coefficients
    reg [2:0] coeff_addr_temp;
    
    initial begin
        state <= START_BYTE_1;
        prev_state <= START_BYTE_1;
        data <= 0;
        unpack_bytes_counter <= 0;
        at_msb <= 1;
        filter_num <= 0;
    end
    
    //State machine for reading in the data received from the RX UART
    always @(posedge RXC) begin
        
        if(RST) begin
            state <= START_BYTE_1;
            data <= 0;
            unpack_bytes_counter <= 0;
            at_msb <= 1;
            filter_num <= 0;
        end
        
        else begin
            data <= UDRRX; //Shift in the data from the receiver module
            
            case(state)
                
                START_BYTE_1: begin //Look for the first start byte
                    unpack_bytes_counter <= 0;
                    at_msb <= 0;
                    filter_num <= filter_num;
                    
                    if(data == 8'hca)
                        state <= START_BYTE_2;
                    else
                        state <= START_BYTE_1;
                end
                
                START_BYTE_2: begin //Look for second start byte
                    unpack_bytes_counter <= 0;
                    at_msb <= 0;
                    filter_num <= filter_num;
                    
                    if(data == 8'hac)
                        state <= FILTER_NUM;
                    else
                        state <= START_BYTE_1;  
                end
                
               FILTER_NUM: begin //Get the filter number
                    unpack_bytes_counter <= 0;
                    at_msb <= 0;
                    
                    if(data >= NUM_FILTERS) begin //Invalid filter number
                        state <= START_BYTE_1;
                        filter_num <= filter_num;
                    end
                    else begin
                        state <= UNPACK_BYTES;
                        filter_num <= data;
                    end
                end
                
                UNPACK_BYTES: begin //Unpack the 16bit coefficients 
                    
                    filter_num <= filter_num;
                    
                    if(unpack_bytes_counter < 5)begin //There are a total of 10 bytes for 5 coefficients
                        at_msb <= ~at_msb;
                        state <= UNPACK_BYTES;
                        if(at_msb) begin //Load msb
                            coeff_mem[unpack_bytes_counter][7:0] <= data; 
                            unpack_bytes_counter <= unpack_bytes_counter + 1;
                            if(unpack_bytes_counter == 4) begin
                                state <= STOP_BYTE;
                            end
                            else begin
                                state <= UNPACK_BYTES;
                            end
                        end
                        else begin //load lsb
                            coeff_mem[unpack_bytes_counter][15:8] <= data;
                            unpack_bytes_counter <= unpack_bytes_counter;
                            state <= UNPACK_BYTES;
                        end
                    end
                    
                    else begin
                        at_msb <= 0;
                        unpack_bytes_counter <= 0;
                        state <= STOP_BYTE;
                    end 
                        
                end
                
                STOP_BYTE: begin //Verify the end of receiving coefficients with correct stop bytes
                    unpack_bytes_counter <= 0;
                    at_msb <= 0;
                    filter_num <= filter_num;
                    
                    if(data == 8'hFA)
                        state <= LOAD_STATE;
                    else
                        state <= START_BYTE_1;

                end
                
                //NOTE, TWO MORE BYTES NEED TO BE SENT VIA UART AFTER STOPBYTE, DOESN'T MATTER WHAT THEY ARE THOUGH
                //This ensures the loading operation can happend
                LOAD_STATE: begin 
                    unpack_bytes_counter <= 0;
                    at_msb <= 0;
                    filter_num <= filter_num;
                    
                    if(LOADING_FINISHED)
                        state <= START_BYTE_1;
                    else
                        state <= LOAD_STATE;
                end
    
            endcase
        end
    
    end
    
    //Detect the change to LOAD_STATE
    always @(posedge SYS_CLK) begin
    
        prev_state <= state;
        if((prev_state != state) & (state == LOAD_STATE))
            START_LOADING <= 1;
        else
            START_LOADING <= 0;
    end
    
    //LOAD the coeffs from memory out synchronous to sys clk. This only should happend when 
    always @(posedge SYS_CLK) begin
        if(RST) begin
            LOADING_FINISHED <= 1'b1;
            coeff_out <= 0;
            coeff_addr_temp <= 0;
            WRITE_EN <= 0;
        end
        else if(START_LOADING) begin  //Initialize loading the coefficients 
            LOADING_FINISHED <= 1'b0;
            coeff_out <= 0;
            coeff_addr_temp <= 0;
            WRITE_EN <= 0;  
        end
        else if(~LOADING_FINISHED) begin //Load the coefficients and set WRITE_EN
            if(coeff_addr_temp < 5) begin
                LOADING_FINISHED <= 1'b0;
                WRITE_EN <= 1'b1;
                coeff_out <= coeff_mem[coeff_addr_temp];
                coeff_addr_temp <= coeff_addr_temp + 1;
            end
            
            else begin //Loading is complete
                LOADING_FINISHED <= 1'b1;
                WRITE_EN <= 1'b0;
                coeff_out <= 0;
                coeff_addr_temp <= 0;
            end
        end
        
        else begin
            LOADING_FINISHED <= 1'b1;
            coeff_out <= 0;
            coeff_addr_temp <= 0;
            WRITE_EN <= 0;
        end
    end
    
    //Register coefficient address so it lines up with the output coefficients
    always @(posedge SYS_CLK) begin
        coeff_addr <= coeff_addr_temp;
    end
endmodule
